// -------------------------------------------------
// CC Nue analysis module
//
// Author: Rory Fitzpatrick
// Updated: 2/19/18
// -------------------------------------------------

#ifndef CCNUE_H
#define CCNUE_H

// Framework includes
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/SubRun.h"
#include "fhiclcpp/ParameterSet.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/View.h"
#include "canvas/Persistency/Common/Ptr.h"
#include "canvas/Persistency/Common/PtrVector.h"
#include "art/Framework/Services/Registry/ServiceHandle.h"
#include "art/Framework/Services/Optional/TFileService.h"
#include "art/Framework/Services/Optional/TFileDirectory.h"
#include "messagefacility/MessageLogger/MessageLogger.h"
#include "canvas/Persistency/Common/FindMany.h"

#include "T962_Objects/MINOS.h"
#include "T962_Objects/Paddles.h"

#include "larcore/Geometry/Geometry.h"
#include "nusimdata/SimulationBase/MCTruth.h"
#include "nusimdata/SimulationBase/MCFlux.h"
#include "lardataobj/Simulation/SimChannel.h"
#include "lardataobj/AnalysisBase/Calorimetry.h"
#include "lardataobj/AnalysisBase/ParticleID.h"
#include "lardataobj/RawData/RawDigit.h"
#include "lardataobj/RawData/BeamInfo.h"
#include "lardata/Utilities/AssociationUtil.h"
#include "lardata/DetectorInfoServices/DetectorPropertiesService.h"
#include "larcoreobj/SummaryData/POTSummary.h"
#include "larsim/MCCheater/BackTrackerService.h"
#include "larsim/MCCheater/ParticleInventoryService.h"
#include "lardataobj/RecoBase/Track.h"
#include "lardataobj/RecoBase/TrackHitMeta.h"
#include "lardataobj/RecoBase/Cluster.h"
#include "lardataobj/RecoBase/Shower.h"
#include "lardataobj/RecoBase/Wire.h"
#include "lardataobj/RecoBase/Hit.h"
#include "lardataobj/RecoBase/EndPoint2D.h"
#include "lardataobj/RecoBase/Vertex.h"
#include "larcoreobj/SimpleTypesAndConstants/geo_types.h"
#include "larreco/Calorimetry/CalorimetryAlg.h"

#include <vector>
#include <cmath> 
#include <string>
#include <sstream>
#include <fstream>
#include <algorithm>
#include <functional>

#include "TTree.h"
#include "TFile.h"
#include "TH1.h"
#include "TSystem.h"
#include "TString.h"
#include "TTimeStamp.h"

const int kMaxTracks     = 1000;  //maximum number of tracks                                                   
//const int kMaxClusters   = 1000;  //maximum number of clusters                                                 
const int kMaxShowers    = 1000;  //maximum number of showers                                                  
//const int kMaxHits       = 20000; //maximum number of hits;                                                    
//const int kMaxPrimaries  = 20000;  //maximum number of primary particles                                       
//const int kMaxTrackHits  = 1000;  //maximum number of hits on a track   

namespace t962 {

  class CCNueSelection : public art::EDAnalyzer {

  public:

    explicit CCNueSelection(fhicl::ParameterSet const& pset);
    virtual ~CCNueSelection();

    void reconfigure(fhicl::ParameterSet const& pset);
    void beginJob();
    void analyze(const art::Event& evt);
    
  protected:

  private: 

    void resetVars();
    bool inFV(double x, double y, double z);
    bool minosMatch(art::Handle< std::vector<recob::Track> > LarTrackHandle, art::Handle< std::vector<t962::MINOS> > MinosTrackHandle, int& nmatched, double& totaldiff, double& best_xdiff, double& best_ydiff, double& best_rdiff, double& best_thetadiff, double& best_dcosx, double& best_dcosy, double& best_dcosz);
    int  endsOnBoundary(art::Ptr<recob::Track> lar_track);
    //    void GetTruthInfo(std::vector<art::Ptr<recob::Hit>> hits, int& origin, int& pdgcode);

    TTree* fTree;
    // run information
    int    run;
    int    subrun;
    int    event;
    double evttime;
    double beamtime;
    double pot;
    double taulife;
    int    isdata;

    //reconstructed information 
    double vtxx_reco;
    double vtxy_reco;
    double vtxz_reco;
    double enu_reco;

    int    ntracks_reco;         //number of reconstructed tracks    
    int    ntracks_onboundary;
    int    trkonboundary[kMaxTracks];

    int    nclusu_reco;
    int    nclusv_reco;

    int    nshws;                      //number of showers
    int    shwid[kMaxShowers];         //recob::Shower::ID()
    float  shwdcosx[kMaxShowers];      //shower direction cosine
    float  shwdcosy[kMaxShowers];
    float  shwdcosz[kMaxShowers];
    float  shwstartx[kMaxShowers];     //shower start position (cm)
    float  shwstarty[kMaxShowers];
    float  shwstartz[kMaxShowers];
    float  shwenergy[kMaxShowers][2];  //shower energy measured on the 3 planes (GeV)
    float  shwdedx[kMaxShowers][2];    //shower dE/dx of the initial track measured on the 3 plane (MeV/cm)
    int    shwbestplane[kMaxShowers];  //recommended plane for energy and dE/dx information
    int    shwg4id[kMaxShowers];       //geant track id for the shower
    
    // MINOS VARIABLES
    int    ntracks_minos;
    int    nmatch;
    double totaldiff;
    double rdiff;
    double xdiff;
    double ydiff;
    double thetadiff;
    double dcosx;
    double dcosy;
    double dcosz;

    int    infidvol;

    //mctruth information 
    int    nuPDG_truth;
    int    ccnc_truth;
    int    mode_truth;
    double enu_truth;
    double nuvtxx_truth;
    double nuvtxy_truth;
    double nuvtxz_truth;

    // parameter set
    std::string fDigitModuleLabel;
    std::string fHitsModuleLabel;
    std::string fLArG4ModuleLabel;
    std::string fCalDataModuleLabel;
    std::string fGenieGenModuleLabel;
    std::string fG4ModuleLabel;
    std::string fClusterModuleLabel;
    std::string fShowerModuleLabel;
    std::string fTrackModuleLabel;
    std::string fEndPoint2DModuleLabel;
    std::string fVertexModuleLabel;
    std::string fMINOSModuleLabel;
    std::string fTrackMatchModuleLabel;
    std::string fPOTModuleLabel;
    std::string fCalorimetryModuleLabel;
    std::string fParticleIDModuleLabel;
    std::string fPaddlesModuleLabel;

    double fvertextrackWindow;
    double fvertexclusterWindow;
    calo::CalorimetryAlg fCalorimetryAlg;

    // MINOS matching
    double fdZ;    //max. z distance between projected T962 track and MINOS track
    double fdXY;   //max. radial distance between projected T962 track and MINOS track
    double fdCosx; //tolerance in dircosX matching
    double fdCosy; //tolerance in dircosY matching
    double fdCosz; //tolerance in dircosZ matching
    double fdBoundary; //distance from a boundary to be considered a track that "ends on a boundary"

  }; // class CCNueSelection

} // t962

// -------------------------------------------------

double lifetimeCorrection(float time) {
  double t = time;

  auto const* detprop = lar::providerFrom<detinfo::DetectorPropertiesService>();

  double timetick = detprop->SamplingRate()*1.e-3; // time smaple in microsec
  double presamplings = detprop->TriggerOffset();

  t -= presamplings;
  time = t * timetick; // in microsec

  double tau = detprop->ElectronLifetime();

  double correction = exp(time/tau);
  return correction;

} // lifetimeCorrection

// -------------------------------------------------

t962::CCNueSelection::CCNueSelection(fhicl::ParameterSet const& pset) :
  EDAnalyzer(pset), 
  fCalorimetryAlg(pset.get<fhicl::ParameterSet>("CalorimetryAlg")) {
  this->reconfigure(pset);
} // CCNueSelection

// -------------------------------------------------

t962::CCNueSelection::~CCNueSelection() {
} // ~CCNueSelection

// -------------------------------------------------

void t962::CCNueSelection::reconfigure(fhicl::ParameterSet const& pset) {
  fDigitModuleLabel = pset.get< std::string >("DigitModuleLabel");
  fHitsModuleLabel = pset.get< std::string >("HitsModuleLabel");
  fLArG4ModuleLabel = pset.get< std::string >("LArGeantModuleLabel");
  fCalDataModuleLabel = pset.get< std::string >("CalDataModuleLabel");
  fGenieGenModuleLabel = pset.get< std::string >("GenieGenModuleLabel");
  fG4ModuleLabel = pset.get< std::string >("G4ModuleLabel");
  fClusterModuleLabel = pset.get< std::string >("ClusterModuleLabel");
  fShowerModuleLabel = pset.get< std::string >("ShowerModuleLabel");
  fTrackModuleLabel = pset.get< std::string >("TrackModuleLabel");
  fEndPoint2DModuleLabel = pset.get< std::string >("EndPoint2DModuleLabel");
  fVertexModuleLabel = pset.get< std::string >("VertexModuleLabel");
  fMINOSModuleLabel = pset.get< std::string >("MINOSModuleLabel");
  fTrackMatchModuleLabel = pset.get< std::string >("TrackMatchModuleLabel");
  fPOTModuleLabel = pset.get< std::string >("POTModuleLabel");
  fCalorimetryModuleLabel = pset.get< std::string >("CalorimetryModuleLabel");
  fParticleIDModuleLabel = pset.get< std::string >("ParticleIDModuleLabel");
  fPaddlesModuleLabel = pset.get< std::string >("PaddlesModuleLabel");
  fvertextrackWindow = pset.get< double >("vertextrackWindow");
  fvertexclusterWindow = pset.get< double >("vertexclusterWindow");
  fdZ = pset.get< double >("dZ");
  fdXY = pset.get< double >("dXY");
  fdCosx = pset.get< double >("dCosx");
  fdCosy = pset.get< double >("dCosy");
  fdCosz = pset.get< double >("dCosz");
  fdBoundary = pset.get< double >("dBoundary");
} // reconfigure

// -------------------------------------------------

void t962::CCNueSelection::beginJob() {
  art::ServiceHandle<art::TFileService> tfs;
  fTree = tfs->make<TTree>("nueselection", "cc nue selection");
  fTree->Branch("run", &run, "run/I");
  fTree->Branch("subrun",&subrun,"subrun/I");
  fTree->Branch("event",&event,"event/I");
  fTree->Branch("evttime",&evttime,"evttime/D");
  fTree->Branch("beamtime",&beamtime,"beamtime/D");
  fTree->Branch("pot",&pot,"pot/D");
  fTree->Branch("taulife",&taulife,"taulife/D");
  fTree->Branch("isdata",&isdata,"isdata/I");

  fTree->Branch("vtxx_reco",&vtxx_reco,"vtxx_reco/D");
  fTree->Branch("vtxy_reco",&vtxy_reco,"vtxy_reco/D");
  fTree->Branch("vtxz_reco",&vtxz_reco,"vtxz_reco/D");
  fTree->Branch("enu_reco",&enu_reco,"enu_reco/D");

  fTree->Branch("ntracks_reco",&ntracks_reco,"ntracks_reco/I");
  fTree->Branch("ntracks_onboundary",&ntracks_onboundary,"ntracks_onboundary/I");
  fTree->Branch("trkonboundary",trkonboundary,"trkonbounary[ntracks_reco]/I");

  fTree->Branch("nclusu_reco",&nclusu_reco,"nclusu_reco/I");
  fTree->Branch("nclusv_reco",&nclusv_reco,"nclusv_reco/I");

  // shower information                                                                                            
  fTree->Branch("nshws",&nshws,"nshws/I");
  fTree->Branch("shwid",shwid,"shwid[nshws]/I");
  fTree->Branch("shwdcosx",shwdcosx,"shwdcosx[nshws]/F");
  fTree->Branch("shwdcosy",shwdcosy,"shwdcosy[nshws]/F");
  fTree->Branch("shwdcosz",shwdcosz,"shwdcosz[nshws]/F");
  fTree->Branch("shwstartx",shwstartx,"shwstartx[nshws]/F");
  fTree->Branch("shwstarty",shwstarty,"shwstarty[nshws]/F");
  fTree->Branch("shwstartz",shwstartz,"shwstartz[nshws]/F");
  fTree->Branch("shwenergy",shwenergy,"shwenergy[nshws][2]/F");
  fTree->Branch("shwdedx",shwdedx,"shwdedx[nshws][2]/F");
  fTree->Branch("shwbestplane",shwbestplane,"shwbestplane[nshws]/I");
  fTree->Branch("shwg4id",shwg4id,"shwg4id[nshws]/I");

  fTree->Branch("ntracks_minos",&ntracks_minos,"ntracks_minos/I");
  fTree->Branch("nmatch",&nmatch,"nmatch/I");
  fTree->Branch("totaldiff",&totaldiff,"totaldiff/D");
  fTree->Branch("rdiff",&rdiff,"rdiff/D");
  fTree->Branch("xdiff",&xdiff,"xdiff/D");
  fTree->Branch("ydiff",&xdiff,"ydiff/D");
  fTree->Branch("thetadiff",&thetadiff,"thetadiff/D");
  fTree->Branch("dcosx",&dcosx,"dcosx/D");  
  fTree->Branch("dcosy",&dcosy,"dcosy/D");
  fTree->Branch("dcosz",&dcosz,"dcosz/D");

  fTree->Branch("infidvol",&infidvol,"infidvol/I");

  fTree->Branch("nuPDG_truth",&nuPDG_truth,"nuPDG_truth/I");
  fTree->Branch("ccnc_truth",&ccnc_truth,"ccnc_truth/I");
  fTree->Branch("mode_truth",&mode_truth,"mode_truth/I");
  fTree->Branch("enu_truth",&enu_truth,"enu_truth/D");
  fTree->Branch("nuvtxx_truth",&nuvtxx_truth,"nuvtxx_truth/D");
  fTree->Branch("nuvtxy_truth",&nuvtxy_truth,"nuvtxy_truth/D");
  fTree->Branch("nuvtxz_truth",&nuvtxz_truth,"nuvtxz_truth/D");

} // beginJob

// -------------------------------------------------

void t962::CCNueSelection::analyze(const art::Event& evt) {

  resetVars();
  
  run = evt.run();
  subrun = evt.subRun();
  event = evt.id().event();

  art::Timestamp ts = evt.time();
  TTimeStamp tts(ts.timeHigh(), ts.timeLow());

  art::Handle<raw::BeamInfo> beam;
  if (evt.getByLabel("beam", beam)) {
    beamtime = (double)beam->get_t_ms();
    beamtime /= 1000.; // in seconds
  }

  if (evt.isRealData()) isdata = 1;
  else isdata = 0;

  // hits, wires, tracks, clusters, showers, vetices, endpoints
  art::Handle< std::vector<recob::Hit> > hitListHandle;
  std::vector<art::Ptr<recob::Hit> > hitlist;
  if (evt.getByLabel(fHitsModuleLabel,hitListHandle))
    art::fill_ptr_vector(hitlist, hitListHandle);

  art::Handle< std::vector<recob::Track> > trackListHandle;
  std::vector<art::Ptr<recob::Track> > tracklist;
  if (evt.getByLabel(fTrackModuleLabel,trackListHandle))
    art::fill_ptr_vector(tracklist, trackListHandle);

  art::Handle< std::vector<recob::EndPoint2D> > endpointListHandle;
  std::vector<art::Ptr<recob::EndPoint2D> > endpointlist;
  if (evt.getByLabel(fEndPoint2DModuleLabel,endpointListHandle))
    art::fill_ptr_vector(endpointlist, endpointListHandle);
  
  art::Handle< std::vector<recob::Vertex> > vertexListHandle;
  std::vector<art::Ptr<recob::Vertex> > vertexlist;
  if (evt.getByLabel(fVertexModuleLabel,vertexListHandle))
    art::fill_ptr_vector(vertexlist, vertexListHandle);

  art::Handle< std::vector<recob::Cluster> > clusterListHandle;
  std::vector<art::Ptr<recob::Cluster> > clusterlist;
  if (evt.getByLabel(fClusterModuleLabel,clusterListHandle))
    art::fill_ptr_vector(clusterlist, clusterListHandle);

  art::Handle< std::vector<recob::Shower> > shwListHandle;
  std::vector<art::Ptr<recob::Shower> > shwlist;
  if (evt.getByLabel(fShowerModuleLabel,shwListHandle))
    art::fill_ptr_vector(shwlist, shwListHandle);

  // minos
  art::Handle< std::vector<t962::MINOS> > minosListHandle;
  std::vector<art::Ptr<t962::MINOS> > minoslist;
  if (evt.getByLabel(fMINOSModuleLabel,minosListHandle))
    art::fill_ptr_vector(minoslist, minosListHandle);
  
  // mc info
  art::Handle< std::vector<simb::MCTruth> > mctruthListHandle;
  std::vector<art::Ptr<simb::MCTruth> > mclist;
  if (evt.getByLabel(fGenieGenModuleLabel,mctruthListHandle))
    art::fill_ptr_vector(mclist, mctruthListHandle);

  // get truth information
  if (!isdata && mclist.size()) {
    art::Ptr<simb::MCTruth> mctruth = mclist[0];
    if (mctruth->NeutrinoSet()) {
      ccnc_truth = mctruth->GetNeutrino().CCNC();
      mode_truth = mctruth->GetNeutrino().Mode();
      nuPDG_truth = mctruth->GetNeutrino().Nu().PdgCode();
      enu_truth = mctruth->GetNeutrino().Nu().E();
      nuvtxx_truth = mctruth->GetNeutrino().Nu().Vx();
      nuvtxy_truth = mctruth->GetNeutrino().Nu().Vy();
      nuvtxz_truth = mctruth->GetNeutrino().Nu().Vz();
    }
  }

  // primitive vertex reco (seriously, it's really primitive)
  if (vertexlist.size()) {
    double vtxxyz[3];
    vertexlist[0]->XYZ(vtxxyz);
    vtxx_reco = vtxxyz[0];
    vtxy_reco = vtxxyz[1];
    vtxz_reco = vtxxyz[2];
  }

  ntracks_reco = tracklist.size();
  ntracks_onboundary = 0;
  for (unsigned int i = 0; i < tracklist.size(); ++i) {
    trkonboundary[i] = 0;
    if (endsOnBoundary(tracklist[i])) {
      trkonboundary[i] = 1;
      ntracks_onboundary++;
    }
  }


  nclusu_reco = 0;
  nclusv_reco = 0;
  // count clusters on each plane
  for (size_t i = 0; i < clusterlist.size(); ++i) {
    switch (clusterlist[i]->View()) {
    case geo::kU :
      nclusu_reco++; break;
    case geo::kV :
      nclusv_reco++; break;
    default : break;
    } // switch view
  }

  // shower info
  if (shwListHandle.isValid()) {
    art::FindManyP<recob::Hit> fmsh(shwListHandle, evt, fShowerModuleLabel);
    nshws = shwlist.size();
    
    for (int i = 0; i < std::min(int(shwlist.size()), kMaxShowers); ++i) {
      shwid[i] = shwlist[i]->ID();
      shwdcosx[i] = shwlist[i]->Direction().X();
      shwdcosy[i] = shwlist[i]->Direction().Y();
      shwdcosz[i] = shwlist[i]->Direction().Z();
      shwstartx[i] = shwlist[i]->ShowerStart().X();
      shwstarty[i] = shwlist[i]->ShowerStart().Y();
      shwstartz[i] = shwlist[i]->ShowerStart().Z();
      for (size_t j = 0; j<(shwlist[i]->Energy()).size(); ++j){
        shwenergy[i][j] = shwlist[i]->Energy()[j];
      }
      for (size_t j = 0; j<(shwlist[i]->dEdx()).size(); ++j){
        shwdedx[i][j] = shwlist[i]->dEdx()[j];
      }
      shwbestplane[i] = shwlist[i]->best_plane();
    }

  } // shower info

  // flags
  bool fInFidVol;
  bool fMinosMatch;
  nmatch = 0;
  fInFidVol = inFV(vtxx_reco, vtxy_reco, vtxz_reco);
  ntracks_minos = minoslist.size();
  fMinosMatch = minosMatch(trackListHandle, minosListHandle, nmatch, totaldiff, xdiff, ydiff, rdiff, thetadiff, dcosx, dcosy, dcosz);

  

  if (fInFidVol && !fMinosMatch) std::cout << "just using some variables" << std::endl; 

  fTree->Fill();

} // analyze

// -------------------------------------------------

void t962::CCNueSelection::resetVars() {

  run = -99999;
  subrun = -99999;
  event = -99999;
  evttime = -99999;
  beamtime = -99999;
  isdata = -99999;
  taulife = -99999;

  vtxx_reco = -99999;
  vtxy_reco  = -99999;
  vtxz_reco  = -99999;
  
  ntracks_reco = -99999;
  ntracks_onboundary = -99999;
  for (int i = 0; i<kMaxTracks; ++i) {
    trkonboundary[i] = -99999;
  }

  nclusu_reco  = -99999;
  nclusv_reco  = -99999;

  nshws = 0;
  for (int i = 0; i<kMaxShowers; ++i){
    shwid[i] = -99999;
    shwdcosx[i] = -99999;
    shwdcosy[i] = -99999;
    shwdcosz[i] = -99999;
    shwstartx[i] = -99999;
    shwstarty[i] = -99999;
    shwstartz[i] = -99999;
    for (int j = 0; j<2; ++j){
      shwenergy[i][j] = -99999;
      shwdedx[i][j] = -99999;
    }
    shwbestplane[i] = -99999;
    shwg4id[i] = -99999;
  }

  ntracks_minos = -99999;
  nmatch = -99999;
  totaldiff = 99999;
  rdiff = -99999;
  xdiff = -99999;
  ydiff = -99999;
  thetadiff = -99999;
  dcosx = -99999;  
  dcosy = -99999;
  dcosz = -99999;

  infidvol = 0;

  nuPDG_truth = -99999;
  ccnc_truth = -99999;
  mode_truth = -99999;
  enu_truth = -99999;
  nuvtxx_truth = -99999;
  nuvtxy_truth = -99999;
  nuvtxz_truth = -99999;

} // resetVars

// -------------------------------------------------

bool t962::CCNueSelection::inFV(double x, double y, double z) {

  geo::GeometryCore const* fGeometry(lar::providerFrom<geo::Geometry>());
  double fDistToEdgeX = fGeometry->DetHalfWidth() - 4;
  double fDistToEdgeY = fGeometry->DetHalfHeight() - 4;
  double fDistToNearZ = -(fGeometry->DetLength() / 2. - 6);
  double fDistToFarZ = fGeometry->DetLength() / 2. - 20; 

  double distInX = x - fGeometry->DetHalfWidth();
  double distInY = y;
  double distInZ = z - 0.5 * fGeometry->DetLength();

  if (std::abs(distInX) < fDistToEdgeX && std::abs(distInY) < fDistToEdgeY && distInZ < fDistToFarZ && distInZ > fDistToNearZ) return true;
  else return false;

} // inFV

// -------------------------------------------------

bool t962::CCNueSelection::minosMatch(art::Handle< std::vector<recob::Track> > LarTrackHandle, art::Handle< std::vector<t962::MINOS> > MinosTrackHandle, int& nmatched, double& min_totaldiff, double& best_xdiff, double& best_ydiff, double& best_rdiff, double& best_thetadiff, double& best_dcosx, double& best_dcosy, double& best_dcosz) {

  bool matchExists = false;

  for (unsigned int i = 0; i < LarTrackHandle->size(); ++i) {
    art::Ptr<recob::Track> lartrack(LarTrackHandle,i);
    int onboundary = endsOnBoundary(lartrack);
    if (!onboundary) continue;

    recob::Track::Point_t larStart, larEnd;
    std::tie(larStart, larEnd) = lartrack->Extent();
    double lardirectionStart[3];
    double lardirectionEnd[3];
    lartrack->Direction(lardirectionStart, lardirectionEnd);

    // check if track needs to be flipped
    if (onboundary == 2) {
      auto tmp = larStart; larStart = larEnd; larEnd = tmp; // flip positions
      for (size_t k = 0; k < 3; ++k) {
	double tmp = lardirectionStart[k];
	lardirectionStart[k] = -lardirectionEnd[k];
	lardirectionEnd[k] = -tmp;
      }
    }
    
    // now loop over minos tracks
    for (unsigned int j = 0; j < MinosTrackHandle->size(); ++j) {
      art::Ptr<t962::MINOS> minostrack(MinosTrackHandle,j);

      if ( (100.0 * minostrack->ftrkVtxZ) > fdZ ) continue; // MINOS track starts too far into detector

      // now define some numbers
      double D = (90. * 0.5) + (42.4 * 2.54) - 5.588; // distance from the front (upstream) of the TPC to the first MINOS plane
      double x_offset = 117.4;
      double y_offset = 19.3;
      if (isdata) {
	x_offset += -3.134 + 1.867576;
	y_offset += -0.837;
      }
      else {
	x_offset += -0.893;
	y_offset += -1.597;
      }

      double dz = D - larEnd.Z() + (100.0 * minostrack->ftrkVtxZ); // z-diff b/w end of t962 trk and begin of minos trk (cm)
      double l = dz/(lardirectionEnd[2]); // 3d distance b/w end of t962 trk and begin of minos trk
      double x_pred = l*lardirectionEnd[0]+larEnd.X(); // predicted x pos of t962 trk at z pos = start of minos trk
      double y_pred = l*lardirectionEnd[1]+larEnd.Y(); // predicted y pos of t962 trk at z pos = start of minos trk
      double dx = 100.0 * minostrack->ftrkVtxX - x_offset - x_pred;
      double dy = 100.0 * minostrack->ftrkVtxY + y_offset - y_pred;
      
      if ( (x_pred + x_offset) > 297.7 || (x_pred + x_offset) < -187.45
	   || (y_pred - y_offset) > 181.71 || (y_pred - y_offset) < -100.29) continue; // prediction outside minos detector

      // cut that you want to use once you optimize them

      /*
      if (sqrt(dx*dx + dy*dy) > fdXY) continue; // predicted coordinates too far from xy pos at start of minos track

      if (std::abs(lardirectionEnd[0] - minostrack->ftrkdcosx) > fdCosx) continue;
      if (std::abs(lardirectionEnd[1] - minostrack->ftrkdcosy) > fdCosx) continue;
      if (std::abs(lardirectionEnd[2] - minostrack->ftrkdcosz) > fdCosx) continue;
      */

      // if you've gotten this far you've got a match that satisfies requirements
      matchExists = true;
      nmatched++;
      
      double rdiff = sqrt(dx*dx + dy*dy);
      double totaldiff = std::abs( rdiff/( (lardirectionEnd[0]*minostrack->ftrkdcosx)+(lardirectionEnd[1]*minostrack->ftrkdcosy)+(lardirectionEnd[2]*minostrack->ftrkdcosz) ) );

      std::cout << min_totaldiff << " " << totaldiff << std::endl;

      if (totaldiff < min_totaldiff) {
	min_totaldiff = totaldiff;
	best_rdiff = rdiff;
	best_xdiff = dx;
	best_ydiff = dy;
	//best_thetadiff = TMath::ACos();
	best_dcosx = std::abs(lardirectionEnd[0] - minostrack->ftrkdcosx);
	best_dcosy = std::abs(lardirectionEnd[1] - minostrack->ftrkdcosy);
	best_dcosz = std::abs(lardirectionEnd[2] - minostrack->ftrkdcosz);
      }

    } // loop over MINOS tracks
  } // loop over t962 tracks
  
  return matchExists;

} // minosMatch

// -------------------------------------------------

int t962::CCNueSelection::endsOnBoundary(art::Ptr<recob::Track> lar_track) {
  
  bool matchstart = false;
  bool matchend = false;

  recob::Track::Point_t larStart, larEnd;
  std::tie(larStart, larEnd) = lar_track->Extent();

  if (std::abs(larEnd.X()) < fdBoundary 
      || std::abs(47. - larEnd.X()) < fdBoundary
      || std::abs(20. + larEnd.Y()) < fdBoundary
      || std::abs(20. - larEnd.Y()) < fdBoundary
      || std::abs(larEnd.Z()) < fdBoundary
      || std::abs(90. - larEnd.Z()) < fdBoundary )
    matchend = true;

  if (std::abs(larStart.X()) < fdBoundary 
      || std::abs(47. - larStart.X()) < fdBoundary
      || std::abs(20. + larStart.Y()) < fdBoundary
      || std::abs(20. - larStart.Y()) < fdBoundary
      || std::abs(larStart.Z()) < fdBoundary
      || std::abs(90. - larStart.Z()) < fdBoundary )
    matchstart = true;

  if (matchend) return 1;
  else if (matchstart) return 2;
  else return 0;

} // endsOnBoundary

// -------------------------------------------------

namespace t962 {

  DEFINE_ART_MODULE(CCNueSelection)

} // t962

#endif
